#ifndef my_plot_stuff
#define my_plot_stuff

#include "stdplot.h"
#include "simple.h"

/* The base class standard_plot just allows fluid text manipulation,
   and it includes defaults for other plot parameters which can be
   manipulated --- all of the following can be called at any time
   during exectuion, and will affect all subsequent printings of the
   plot:

   void set_top_margin(double inches)
   void set_bottom_margin(double inches)
   void set_left_margin(double inches)
   void set_right_margin(double inches)
   void reset_margin_defaults()

   void suppress_legend()
   void request_same_page_legend()
   void request_separate_legend()
   void request_big_picture_too()
   void suppress_big_picture()

   void suppress_x_axis_labels()
   void enable_x_axis_labels()
   void suppress_x_axis_tics()
   void enable_x_axis_tics()
   void suppress_x_axis_minor_tics()
   void enable_x_axis_minor_tics()
   void suppress_y_axis_labels()
   void enable_y_axis_labels()
   void suppress_y_axis_tics()
   void enable_y_axis_tics()
   void suppress_y_axis_minor_tics()
   void enable_y_axis_minor_tics()
   void suppress_z_axis_labels()
   void enable_z_axis_labels()
   void suppress_z_axis_tics()
   void enable_z_axis_tics()
   void suppress_z_axis_minor_tics()
   void enable_z_axis_minor_tics()
   void suppress_all_tics()
   void enable_all_tics()
   void suppress_all_minor_tics()
   void enable_all_minor_tics()
   void set_x_axis_label_font(const char * font_name,
                              double font_size)
   void set_x_axis_label_style(value_format style_request)
   void set_y_axis_label_font(const char * font_name,
                              double font_size)
   void set_y_axis_label_style(value_format style_request)
   void set_z_axis_label_font(const char * font_name,
                              double font_size)
   void set_z_axis_label_style(value_format style_request)


   These will only *add* text to the plot, if you want to *change*
   text, they return a pointer to the text string (and you provide
   the code to manipulate existing text).  NOTE, the current value is
   the ONLY pointer to dynamically allocated string storage.  It
   should be freed before being changed AND the routine will continue
   to treat the pointer as dynamically allocated memory; so, you
   can expect the pointer to be freed at termination.

   There is no mechanism for changing the font of existing text,
   nor is there a mechanism for *removing* text.  The idea was to
   encourage verbosity; however, there are practical limits to the
   amount you can include and still see a reasonable image.

   char ** set_plot_title(const char * text_string,
                          const char * font_name,
                          double       font_size)
   char ** set_footnote(const char * text_string,
                        const char * font_name,
                        double       font_size)
   char ** set_plot_subtitle(const char * text_string,
                             const char * font_name,
                             double       font_size)
   char ** set_plot_subsubtitle(const char * text_string,
                                const char * font_name,
                                double       font_size)
   char ** set_legend_info(const char * text_string,
                           const char * font_name,
                           double       font_size)
   char ** set_legend_subinfo(const char * text_string,
                              const char * font_name,
                              double       font_size)
   char ** set_x_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   char ** set_y_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   char ** set_z_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   char ** set_plot_title(const char * text_string)
   char ** set_footnote(const char * text_string)
   char ** set_plot_subtitle(const char * text_string)
   char ** set_plot_subsubtitle(const char * text_string)
   char ** set_x_axis_description(const char * x_axis_description)
   char ** set_y_axis_description(const char * y_axis_description)
   char ** set_z_axis_description(const char * z_axis_description)
   char ** set_legend_info(const char * legend_info)
   char ** set_legend_subinfo(const char * legend_subinfo)

   These are special --- they must be set exactly once for the
   certain plot classes and precede a call to options_complete.  

   For the scarce_histogram plot, these have no affect (it figures out the 
   range of data by itself). 

   void set_x_range(UInt64 user_min, UInt64 user_max)
   void set_y_range(UInt64 user_min, UInt64 user_max)
*/

class density_plot : public standard_plot
{
   UInt32 * storage;

protected:

   void prepare_print()
   {
      UInt32 z_max = 0;
      UInt32 value;
      for (UInt32 loop1 = 0; loop1 <= scale_x.storage_max; loop1++)
      {
         for (UInt32 loop2 = 0; loop2 <= scale_y.storage_max; loop2++)
         {
            value = * storage_location_index(loop1, loop2);
            if (value > z_max)
            {
               z_max = value;
            };
         };
      };
      set_z_range(z_max);
   };

public:
   void options_complete()
   {
      if ((storage = (UInt32 *)calloc((scale_x.storage_max + 1)*
                                      (scale_y.storage_max + 1),
                                      sizeof(UInt32))) == NULL)
      {
         fprintf(stderr, "Unable to allocate memory for density storage\n");
         fprintf(stderr, "Aborting\n");
         exit(1);
      };
   };

   UInt32 * storage_location_index(UInt32 x, UInt32 y)
   {
      return storage + (x + y * (scale_x.storage_max + 1));
   };

   UInt32 * storage_location_user(UInt64 x, UInt64 y)
   {
      if ((x >= scale_x.user_min) && (x <= scale_x.user_max) &&
          (y >= scale_y.user_min) && (y <= scale_y.user_max))
      {
         return storage_location_index(storage_location(scale_x, x),
                                       storage_location(scale_y, y));
      }
      else
      {
         return NULL;
      }
   };

   void point_acculmulate(UInt64 x, UInt64 y)
   {
      UInt32 * data_ptr = storage_location_user(x, y);
      if (data_ptr != NULL)
      {
         (*data_ptr)++;
      };
   };

   void point_set(UInt64 x, UInt64 y, UInt32 value)
   {
      UInt32 * data_ptr = storage_location_user(x, y);
      if (data_ptr != NULL)
      {
         (*data_ptr) = value;
      };
   };

   void emit_image(FILE * output_stream,
                   double left_x_position,
                   double right_x_position,
                   double upper_y_position,
                   double lower_y_position)
   {
      double size_x = right_x_position - left_x_position;
      double size_y = upper_y_position - lower_y_position;
      fprintf(output_stream, "%% Plot image\n");
      fprintf(output_stream, "gsave %g %g translate\n",
                             left_x_position,
                             lower_y_position);
      fprintf(output_stream, "      1 1 scale\n");
      fprintf(output_stream, "      %u %u 8\n",
                             scale_x.storage_max + 1,
                             scale_y.storage_max + 1);
      fprintf(output_stream, "      [ %g 0 0 %g 0 0 ]\n",
                             (scale_x.storage_max + 1.0) / size_x,
                             (scale_y.storage_max + 1.0) / size_y);
      fprintf(output_stream, "      {<\n      ");

      UInt32 line_data_count = 0;
      UInt32 value;
      for (UInt32 loop1 = 0; loop1 <= scale_y.storage_max; loop1++)
      {
         for (UInt32 loop2 = 0; loop2 <= scale_x.storage_max; loop2++)
         {
            if (line_data_count == 20)
            {
               fprintf(output_stream, "\n      ");
               line_data_count = 0;
            };
            value = * storage_location_index(loop2, loop1);
            byte_print(output_stream,
                       (UChar)storage_location_method_two(scale_z, value));
            line_data_count++;
         };
         fprintf(output_stream, "\n      ");
         line_data_count = 0;
      };
      fprintf(output_stream, "      >}\n");
      fprintf(output_stream, "      image\n");
      fprintf(output_stream, "grestore\n");
   };

   void kill()
   {
      kill_standard_plot();
      free(storage);
   };
};

enum histogram_type
{
   STANDARD_HIST,
   LEFT_INTEGRATED_HIST,
   RIGHT_INTEGRATED_HIST
};

class histogram_plot : public standard_plot
{
protected:
   histogram_type hist_type;
   double greyscale;
   UInt32 * storage;

public:
   void options_complete()
   {
      if ((storage = (UInt32 *)calloc(scale_x.storage_max + 1,
                                      sizeof(UInt32))) == NULL)
      {
         fprintf(stderr, "Unable to allocate memory for histogram storage\n");
         fprintf(stderr, "Aborting\n");
         exit(1);
      };
   };

protected:
   UInt32 * storage_location_index(UInt32 x)
   {
      return storage + x;
   };

   UInt32 * storage_location_user(UInt64 x)
   {
      if ((x >= scale_x.user_min) && (x <= scale_x.user_max))
      {
         return storage_location_index(storage_location(scale_x, x));
      }
      else
      {
         return NULL;
      }
   };

public:
   void point_acculmulate(UInt64 x)
   {
      UInt32 * data_ptr = storage_location_user(x);
      if (data_ptr != NULL)
      {
         (*data_ptr)++;
      };
   };

   void point_set(UInt64 x, UInt32 value)
   {
      UInt32 * data_ptr = storage_location_user(x);
      if (data_ptr != NULL)
      {
         (*data_ptr) = value;
      };
   };
/*
   Let user play with the type of histogram
*/
   void set_histogram_type(histogram_type requested_hist_type)
   {
      hist_type = requested_hist_type;
   };

protected:
   void emit_image(FILE * output_stream, double ly)
   {
      UInt32 data;
      for (UInt32 loop = 0; loop <= scale_x.storage_max; loop++)
      {
         data = * storage_location_index(loop);
         greyscale = 1.0 - greyscale;
         if (data != 0)
         {
            double lx = position(box_edge_first_from_storage(scale_x, loop));
            double rx = position(box_edge_second_from_storage(scale_x, loop));
            double uy = position(box_center(scale_y, data));
            emit_block(output_stream, lx, rx, uy, ly, greyscale);
         };
      };
   };

   void emit_image_left_integrate(FILE * output_stream, double ly)
   {
      UInt32 data = 0;
      UInt32 loop;
      for (loop = scale_x.storage_max; loop > 0; loop--)
      {
         data += * storage_location_index(loop);
         greyscale = 1.0 - greyscale;
         if (data != 0)
         {
            double lx = position(box_edge_first_from_storage(scale_x, loop));
            double rx = position(box_edge_second_from_storage(scale_x, loop));
            double uy = position(box_center(scale_y, data));
            emit_block(output_stream, lx, rx, uy, ly, greyscale);
         };
      };
      data += * storage_location_index(0);
      greyscale = 1.0 - greyscale;
      if (data != 0)
      {
         double lx = position(box_edge_first_from_storage(scale_x, loop));
         double rx = position(box_edge_second_from_storage(scale_x, loop));
         double uy = position(box_center(scale_y, data));
         emit_block(output_stream, lx, rx, uy, ly, greyscale);
      };
   };

   void emit_image_right_integrate(FILE * output_stream, double ly)
   {
      UInt32 data = 0;
      for (UInt32 loop = 0; loop <= scale_x.storage_max; loop++)
      {
         data = * storage_location_index(loop);
         greyscale = 1.0 - greyscale;
         if (data != 0)
         {
            double lx = position(box_edge_first_from_storage(scale_x, loop));
            double rx = position(box_edge_second_from_storage(scale_x, loop));
            double uy = position(box_center(scale_y, data));
            emit_block(output_stream, lx, rx, uy, ly, greyscale);
         };
      };
   };

   void emit_image(FILE * output_stream,
                   double left_x_position,
                   double right_x_position,
                   double upper_y_position,
                   double lower_y_position)
   {
      greyscale = 0.25;
      fprintf(output_stream, "%% Histogram Image\n");
      switch (hist_type)
      {
         case STANDARD_HIST:
         {
            emit_image(output_stream, lower_y_position);
         }; break;
         case LEFT_INTEGRATED_HIST:
         {
            UInt64 data = 0;
            emit_image_left_integrate(output_stream, lower_y_position);
         }; break;
         case RIGHT_INTEGRATED_HIST:
         {
            UInt64 data = 0;
            emit_image_right_integrate(output_stream, lower_y_position);
         };
      };
   };

   void prepare_print()
   {
      UInt32 max_count = 0;
      UInt32 loop;
      if (hist_type == STANDARD_HIST)
      {
         for (loop = 0; loop <= scale_x.storage_max; loop++)
         {
            UInt32 data = * storage_location_index(loop);
            if (max_count < data)
            {
               max_count = data;
            };
         };
      }
      else
      {
         for (loop = 0; loop <= scale_x.storage_max; loop++)
         {
            max_count += * storage_location_index(loop);
         };
      };
      fprintf(stderr, "About to set y range 0, %d\n", max_count);
      set_y_range(0, max_count);
   };

public:
   histogram_plot()
   {
      suppress_legend();
      hist_type = STANDARD_HIST;
   };

   void kill()
   {
      kill_standard_plot();
      free(storage);
   };
};

/*
class scarce_histogram_plot : public standard_plot
{
protected:
   histogram_type hist_type;
   double greyscale;
   AVL_tree<UInt64, UInt64> Set;

public:
   void set_histogram_type(histogram_type requested_hist_type)
   {
      hist_type = requested_hist_type;
   };

protected:
   void emit_image(FILE * output_stream,
	      AVL_tree<UInt64, UInt64>::AVL_tree_node * temp_root,
	      double ly)
   {
      if (temp_root != NULL)
      {
         emit_image(output_stream, temp_root->left_child, ly);
         greyscale = 1.0 - greyscale;
         double lx = position(box_edge_first(scale_x, temp_root->key));
         double rx = position(box_edge_second(scale_x, temp_root->key));
         double uy = position(box_center(scale_y, temp_root->data));
         emit_block(output_stream, lx, rx, uy, ly, greyscale);
         emit_image(output_stream, temp_root->right_child, ly);
      };
   };

   void emit_image_right_integrate(FILE * output_stream,
                                   AVL_tree<UInt64,UInt64>::AVL_tree_node * 
                                      temp_root,
	                           double ly,
                                   UInt64 & data)
   {
      if (temp_root != NULL)
      {
         emit_image_right_integrate(output_stream,
                                    temp_root->left_child,
                                    ly, data);
         data += temp_root->data;
         greyscale = 1.0 - greyscale;
         double lx = position(box_edge_first(scale_x, temp_root->key));
         double rx = position(box_edge_second(scale_x, temp_root->key));
         double uy = position(box_center(scale_y, data));
         emit_block(output_stream, lx, rx, uy, ly, greyscale);
         emit_image_right_integrate(output_stream,
                                    temp_root->right_child, ly, data);
      };
   };

   void emit_image_left_integrate(FILE * output_stream,
                                  AVL_tree<UInt64, UInt64>::AVL_tree_node * 
                                     temp_root,
                                  double ly, UInt64 & data)
   {
      if (temp_root != NULL)
      {
         emit_image_left_integrate(output_stream,
                                    temp_root->right_child,
                                    ly, data);
         data += temp_root->data;
         greyscale = 1.0 - greyscale;
         double lx = position(box_edge_first(scale_x, temp_root->key));
         double rx = position(box_edge_second(scale_x, temp_root->key));
         double uy = position(box_center(scale_y, data));
         emit_block(output_stream, lx, rx, uy, ly, greyscale);
         emit_image_left_integrate(output_stream,
                                    temp_root->left_child, ly, data);
      };
   };

   void emit_image(FILE * output_stream,
                   double left_x_position,
                   double right_x_position,
                   double upper_y_position,
                   double lower_y_position)
   {
      greyscale = 0.25;
      fprintf(output_stream, "%% Histogram Image\n");
      switch (hist_type)
      {
         case STANDARD_HIST:
         {
            emit_image(output_stream, Set.root, lower_y_position);
         }; break;
         case LEFT_INTEGRATED_HIST:
         {
            UInt64 data = 0;
            emit_image_left_integrate(output_stream,
                                      Set.root, lower_y_position, data);
         }; break;
         case RIGHT_INTEGRATED_HIST:
         {
            UInt64 data = 0;
            emit_image_right_integrate(output_stream,
                                       Set.root, lower_y_position, data);
         };
      };
   };

   void prepare_print()
   {
      UInt64 min_value;
      UInt64 max_value;
      UInt64 min_count;
      UInt64 max_count;
      if (Set.find_key_extremes(min_value, max_value) != SUCCESS)
      {
	 fprintf(stderr, "Severe error --- empty histogram plot.\n");
	 fprintf(stderr, "Aborting\n");
	 exit(1);
      }
      else
      {
         set_x_range(min_value, max_value);
      };
      if (hist_type == STANDARD_HIST)
      {
         if (Set.find_data_extremes(min_count, max_count) != SUCCESS)
         {
            fprintf(stderr, "Severe error --- empty histogram plot.\n");
            fprintf(stderr, "Possible corruption in data structure.\n");
            fprintf(stderr, "Aborting\n");
	    exit(1);
         };
         fprintf(stderr, "Data extremes are %u %u \n", min_count, max_count);
      }
      else
      {
         min_count = 0;
         if ((max_count = Set.sum_of_entries()) == 0)
         {
            fprintf(stderr, "Severe error --- empty histogram plot.\n");
            fprintf(stderr, "Possible corruption in data structure.\n");
            fprintf(stderr, "Aborting\n");
	    exit(1);
         };
         fprintf(stderr, "Sumof entries is %u \n", max_count);
      };
      set_y_range(min_count, max_count);
   };

public:
   scarce_histogram_plot()
   {
      suppress_legend();
      hist_type = STANDARD_HIST;
      Set.reset();
   };

   void point_acculmulate(UInt64 value)
   {
      Set.increment(value);
   };

   void point_set(UInt64 value1, UInt32 value2)
   {
      Set.insert_entry(value1, value2);
   };

   void options_complete()
   {
   };

   void kill()
   {
      kill_standard_plot();
      Set.kill();
   };
};
*/
#endif

