#ifndef dvw_simple_structures_header
#define dvw_simple_structures_header
 
#include <stdlib.h>
#include <stdio.h>
#include <memory.h>
#include <string.h>
#include "general.h"
/* ***************************************************************
   ***************************************************************
   **                                                           **
   ** Simple Multi-purpose Structures                           **
   **                                                           **
   ** Author: Douglas Van Wieren                                **
   ** Date:  April 8, 1995                                      **
   **                                                           **
   ** Description:                                              **
   **                                                           **
   **      This header file contains some very simple           **
   **      structures                                           **
   **                                                           **
   ***************************************************************
   *************************************************************** */
 

template <class keytype, class datatype>
class AVL_tree
{
protected:
   class AVL_tree_node
   {
   public:
      keytype           key;
      datatype           data;
      AVL_tree_node *  right_child;
      AVL_tree_node *  left_child;
      AVL_status       node_avl_status;
   };
   AVL_tree_node * root;
   UInt32 size;
 
   void rotate_right(AVL_tree_node * & temp_root)
   {
      AVL_tree_node * save_root = temp_root;
      temp_root              = temp_root->left_child;
      save_root->left_child   = temp_root->right_child;
      temp_root->right_child  = save_root;
   };
   void rotate_left(AVL_tree_node * & temp_root)
   {
      AVL_tree_node * save_root = temp_root;
      temp_root              = temp_root->right_child;
      save_root->right_child  = temp_root->left_child;
      temp_root->left_child   = save_root;
   };
   /* the following two routines assume a valid AVL tree and
      return a valid AVL tree of the same height without "heaviness"
      on a particular side. */
   void avoid_left_heavy(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == LEFT_TALLER)
      {
         avoid_right_heavy(temp_root->left_child);
         rotate_right(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status = RIGHT_TALLER;
            (temp_root->right_child)->node_avl_status = BALANCED;
         }
         else
         {
            temp_root->node_avl_status = RIGHT_TALLER;
            (temp_root->right_child)->node_avl_status = RIGHT_TALLER;
         }
      };
   };
   void avoid_right_heavy(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == RIGHT_TALLER)
      {
         avoid_left_heavy(temp_root->right_child);
         rotate_left(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status = LEFT_TALLER;
            (temp_root->left_child)->node_avl_status = BALANCED;
         }
         else
         {
            temp_root->node_avl_status = LEFT_TALLER;
            (temp_root->left_child)->node_avl_status = LEFT_TALLER;
         }
      };
   };
 
   /* these routines return a flag if the tree has grown in height */
   change_flag left_child_increased(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == LEFT_TALLER)
      {
         avoid_right_heavy(temp_root->left_child);
         rotate_right(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status               = RIGHT_TALLER;
            (temp_root->right_child)->node_avl_status = LEFT_TALLER;
            return INCREASED;
         }
         else /* (temp_root == LEFT_TALLER) */
         {
            temp_root->node_avl_status               = BALANCED;
            (temp_root)->right_child->node_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_root->node_avl_status == BALANCED)
      {
         temp_root->node_avl_status = LEFT_TALLER;
         return INCREASED;
      }
      else /* (temp_root->node_avl_status == RIGHT_TALLER) */
      {
         temp_root->node_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
   change_flag right_child_increased(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == RIGHT_TALLER)
      {
         avoid_left_heavy(temp_root->right_child);
         rotate_left(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status              = LEFT_TALLER;
            (temp_root->left_child)->node_avl_status = RIGHT_TALLER;
            return INCREASED;
         }
         else /* (temp_root == RIGHT_TALLER) */
         {
            temp_root->node_avl_status              = BALANCED;
            (temp_root->left_child)->node_avl_status = BALANCED;
            return UNCHANGED;
         };
      }
      else if (temp_root->node_avl_status == BALANCED)
      {
         temp_root->node_avl_status = RIGHT_TALLER;
         return INCREASED;
      }
      else /* (temp_root->node_avl_status == LEFT_TALLER) */
      {
         temp_root->node_avl_status = BALANCED;
         return UNCHANGED;
      };
   };
 
   change_flag left_child_decreased(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == RIGHT_TALLER)
      {
         avoid_left_heavy(temp_root->right_child);
         rotate_left(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status              = LEFT_TALLER;
            (temp_root->left_child)->node_avl_status = RIGHT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_root == RIGHT_TALLER) */
         {
            temp_root->node_avl_status              = BALANCED;
            (temp_root->left_child)->node_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_root->node_avl_status == BALANCED)
      {
         temp_root->node_avl_status = RIGHT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_root->node_avl_status == LEFT_TALLER) */
      {
         temp_root->node_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
   change_flag right_child_decreased(AVL_tree_node * & temp_root)
   {
      if (temp_root->node_avl_status == LEFT_TALLER)
      {
         avoid_right_heavy(temp_root->left_child);
         rotate_right(temp_root);
         if (temp_root->node_avl_status == BALANCED)
         {
            temp_root->node_avl_status               = RIGHT_TALLER;
            (temp_root->right_child)->node_avl_status = LEFT_TALLER;
            return UNCHANGED;
         }
         else /* (temp_root == LEFT_TALLER) */
         {
            temp_root->node_avl_status               = BALANCED;
            (temp_root->right_child)->node_avl_status = BALANCED;
            return DECREASED;
         };
      }
      else if (temp_root->node_avl_status == BALANCED)
      {
         temp_root->node_avl_status = LEFT_TALLER;
         return UNCHANGED;
      }
      else /* (temp_root->node_avl_status == RIGHT_TALLER) */
      {
         temp_root->node_avl_status = BALANCED;
         return DECREASED;
      };
   };
 
   change_flag insert_tree(AVL_tree_node * & temp_root, keytype key, datatype data)
   {
      if (temp_root == NULL)
      {
         frail_allocate(temp_root, AVL_tree_node);
         size++;
         temp_root->key             = key;
         temp_root->data            = data;
         temp_root->right_child     = NULL;
         temp_root->left_child      = NULL;
         temp_root->node_avl_status = BALANCED;
         return INCREASED;
      }
      else if (key < temp_root->key)
      {
         if (insert_tree(temp_root->left_child, key, data) == INCREASED)
         {
            return left_child_increased(temp_root);
         };
      }
      else if (key > temp_root->key)
      {
         if (insert_tree(temp_root->right_child, key, data) == INCREASED)
         {
            return right_child_increased(temp_root);
         }
      }
      else /* (key == temp_root->key) */
      {
         temp_root->data = data;
      };
      return UNCHANGED;
   };
 
   change_flag increment_data(AVL_tree_node * & temp_root, keytype key)
   {
      if (temp_root == NULL)
      {
         frail_allocate(temp_root, AVL_tree_node);
         size++;
         temp_root->key             = key;
         temp_root->data            = 1;
         temp_root->right_child     = NULL;
         temp_root->left_child      = NULL;
         temp_root->node_avl_status = BALANCED;
         return INCREASED;
      }
      else if (key < temp_root->key)
      {
         if (increment_data(temp_root->left_child, key) == INCREASED)
         {
            return left_child_increased(temp_root);
         };
      }
      else if (key > temp_root->key)
      {
         if (increment_data(temp_root->right_child, key) == INCREASED)
         {
            return right_child_increased(temp_root);
         }
      }
      else /* (key == temp_root->key) */
      {
         temp_root->data++;
      };
      return UNCHANGED;
   };
 
   AVL_tree_node * find(AVL_tree_node * temp_root, keytype key)
   {
      if (temp_root == NULL)
      {
         return NULL;
      }
      else
      {
         if (key < temp_root->key)
         {
            return find(temp_root->left_child, key);
         }
         else if (key == temp_root->key)
         {
            return temp_root;
         }
         else /* (key > temp_root->key) */
         {
            return find(temp_root->right_child, key);
         };
      };
   };
 
   AVL_tree_node * find_leftmost(AVL_tree_node * temp_root)
   {
      if ((temp_root != NULL) && (temp_root->left_child != NULL))
      {
         return find_leftmost(temp_root->left_child);
      }
      else
      {
         return temp_root;
      };
   };
 
   AVL_tree_node * find_rightmost(AVL_tree_node * temp_root)
   {
      if ((temp_root != NULL) && (temp_root->right_child != NULL))
      {
         return find_rightmost(temp_root->right_child);
      }
      else
      {
         return temp_root;
      };
   };
 
   change_flag extract_root(AVL_tree_node * & temp_root, keytype key)
   {
      if ((temp_root->left_child  == NULL) &&
          (temp_root->right_child == NULL))
      {
         free(temp_root);
         size--;
         temp_root = NULL;
         return DECREASED;
      }
      else if (temp_root->node_avl_status == LEFT_TALLER)
      {
         AVL_tree_node * condemned = find_rightmost(temp_root->left_child);
         swap(condemned->key,  temp_root->key);
         swap(condemned->data, temp_root->data);
         if (extract_tree(temp_root->left_child, key) == DECREASED)
         {
            return left_child_decreased(temp_root);
         };
      }
      else
      {
         AVL_tree_node * condemned = find_leftmost(temp_root->right_child);
         swap(condemned->key,  temp_root->key);
         swap(condemned->data, temp_root->data);
         if (extract_tree(temp_root->right_child, key) == DECREASED)
         {
            return right_child_decreased(temp_root);
         };
      };
      return UNCHANGED;
   };
 
   change_flag extract_tree(AVL_tree_node * & temp_root, keytype key)
   {
      if (key == temp_root->key)
      {
         extract_root(temp_root, key);
      }
      else if (key < temp_root->key)
      {
         if (extract_tree(temp_root->left_child, key) == DECREASED)
         {
            return left_child_decreased(temp_root);
         };
      }
      else /* (key >temp_root->key) */
      {
         if (extract_tree(temp_root->right_child, key) == DECREASED)
         {
            return right_child_decreased(temp_root);
         };
      }
      return UNCHANGED;
   };

   void print_tree(AVL_tree_node * temp_root, FILE * stream)
   {
      if (temp_root != NULL)
      {
         print_tree(temp_root->left_child, stream);
         fprintf(stream, "%ld %ld\n", temp_root->key, temp_root->data);
         print_tree(temp_root->right_child, stream);
      };
   };

   void find_data_extremes(AVL_tree_node * temp_root,
                           datatype &  min_data,
                           datatype &  max_data)
   {
      if (temp_root != NULL)
      {
         if (min_data > temp_root->data) min_data = temp_root->data;
         if (max_data < temp_root->data) max_data = temp_root->data;
         find_data_extremes(temp_root->left_child, min_data, max_data);
         find_data_extremes(temp_root->right_child, min_data, max_data);
      };
   };

   datatype sum_of_entries(AVL_tree_node * temp_root)
   {
      if (temp_root != NULL)
      {
         return temp_root->data + sum_of_entries(temp_root->left_child)
                                + sum_of_entries(temp_root->right_child);
      }
      else
      {
         return (datatype)0;
      };
   };

   void kill_tree(AVL_tree_node * & temp_root)
   {
      if (temp_root != NULL)
      {
         kill_tree(temp_root->left_child);
         kill_tree(temp_root->right_child);
         free(temp_root);
         size--;
         temp_root = NULL;
      };
   };
 
public:
 
   int avl_tree()
   {
      fprintf(stderr, "AVL tree constructor called \n");
      root = NULL;
      size = 0;
      return 0;
   };
   void reset()
   {
      root = NULL;
      size = 0;
   };
   void insert_entry(keytype key, datatype data)
   {
      insert_tree(root, key, data);
   };
   void increment(keytype key)
   {
      increment_data(root, key);
   };

   datatype * locate_address_of_data(keytype key)
   {
      AVL_tree_node * temp = find(root, key);
      if (temp == NULL)
      {
         return NULL;
      }
      else
      {
         return (& (temp->data));
      };
   };

   void print(FILE * stream)
   {
      print_tree(root, stream);
   };

   void extract_entry(keytype key)
   {
      extract_tree(root, key);
   };

   datatype sum_of_entries()
   {
      return sum_of_entries(root);
   };

   attempt_flag find_key_extremes(keytype &  min_key,
                                  keytype &  max_key)
   {
      if (root != NULL)
      {
         AVL_tree_node * temp_root = root;
         while (temp_root->left_child != NULL)
         {
            temp_root = temp_root->left_child;
         };
         min_key = temp_root->key;
         temp_root = root;
         while (temp_root->right_child != NULL)
         {
            temp_root = temp_root->right_child;
         };
         max_key = temp_root->key;
         return SUCCESS;
      }
      else
      {
         return FAILURE;
      };
   };

   attempt_flag find_data_extremes(datatype &  min_data,
                                   datatype &  max_data)
   {
      if (root == NULL)
      {
         return FAILURE;
      }
      else
      {
         min_data = root->data;
         max_data = root->data;
         find_data_extremes(root->left_child, min_data, max_data);
         find_data_extremes(root->right_child, min_data, max_data);
         return SUCCESS;
      };
   };

   void kill()
   {
      kill_tree(root);
   };
   friend class scarce_histogram_plot;
};


#endif
