#ifndef my_standard_plot_stuff
#define my_standard_plot_stuff
 
#include "plotbase.h"

#define FONT_FRACTION (1.0)
#define INTERLINE_FRACTION (0.5)
#define INVERT_IMAGE
/*
    This class implements tons of junk to allow the user to place text
    items on the screen.  It includes an optional legend area where
    the classes developed from this will a special, separate image which, 
    in some mysterious way, describes the plot image.
 
    All the code tricks are designed to make this flexible; so, after
    some unexplained pictures, we can finally explain the wierd stuff
    in plotbase.h.
 
 
 
    General plot layout:  Margins with and without legend
 
 
                                 11 inches
     _____________________________________________________________________
    |       ^                                                             |
    |       |                                                             |
    |       | Y MARGIN TOP                                                |
    |       |                                                             |
    |       v                                                             |
    |<----->._____________________________________________________        |
    |       |                                                     |       |
    |   X   |                                                     |       |
    |       |                                                     |       |
    |   M   |                                                     |       |
    |   A   |                                                     |       |
    |   R   |                                                     |       |
 8  |   G   |                                                     |       |
1/2 |   I   |                                                     |       |
    |   N   |                                                     |       |
 i  |       |                                                     |       |
 n  |   L   |                                                     |       |
 c  |   E   |                   Image Area                        |   X   |
 h  |   F   |                                                     |       |
 e  |   T   |                                                     |   M   |
 s  |       |                                                     |   A   |
    |       |                                                     |   R   |
    |       |                                                     |   G   |
    |       |                                                     |   I   |
    |       |                                                     |   N   |
    |       |                                                     |       |
    |       |                                                     |   R   |
    |       |                                                     |   I   |
    |       |                                                     |   G   |
    |       |                                                     |   H   |
    |       |                                                     |   T   |
    |       |                                                     |       |
    |       |_____________________________________________________.<----->|
    |                                                             ^       |
    |                                                             |       |
    |                                             Y MARGIN BOTTOM |       |
    |                                                             |       |
    |_____________________________________________________________v_______|
(0,0)
 
 
 
 
 
                                 11 inches
     _____________________________________________________________________
    |       ^                                                             |
    |       |                                                             |
    |       | Y MARGIN TOP                                                |
    |       |                                                             |
    |       v                                                             |
    |<----->.______________________________         ______________        |
    |       |                              |       |              |       |
    |   X   |                              |       |              |       |
    |       |                              |       |              |       |
    |   M   |                              |       |              |       |
    |   A   |                              |   X   |              |       |
    |   R   |                              |       |              |       |
 8  |   G   |                              |   M   |              |       |
1/2 |   I   |                              |   A   |              |       |
    |   N   |                              |   R   |              |       |
 i  |       |                              |   G   |              |       |
 n  |   L   |                              |   I   |              |       |
 c  |   E   |                              |   N   |    LEGEND    |   X   |
 h  |   F   |        IMAGE AREA            |<----->|     AREA     |       |
 e  |   T   |                              |   I   |              |   M   |
 s  |       |                              |   N   |              |   A   |
    |       |                              |   T   |              |   R   |
    |       |                              |   E   |              |   G   |
    |       |                              |   R   |              |   I   |
    |       |                              |   N   |              |   N   |
    |       |                              |   A   |              |       |
    |       |                              |   L   |    LEGEND    |   R   |
    |       |                              |       |<---X SIZE--->|   I   |
    |       |                              |       |              |   G   |
    |       |                              |       |              |   H   |
    |       |                              |       |              |   T   |
    |       |                              |       |              |       |
    |       |______________________________|       |______________.<----->|
    |                                                             ^       |
    |                                                             |       |
    |                                             Y MARGIN BOTTOM |       |
    |                                                             |       |
    |_____________________________________________________________v_______|
 

 
    General Plot Text Layout
 
     _____________________________________________________________________
    |                                                                     |
    |                                                                     |
    |                                                                     |
    |                                                                     |
    |                                                                     |
    |                           "Plot Title"                              |
    |                                                                     |
    |                                                                     |
    |        "Plot Subtitle"                          "Legend Info"       |
    |                                                                     |
    |                                                                     |
    |                                                                     |
    |   Y                                           Z                     |
    |   |                                           |                     |
    |   A  Y                                        A  Z                  |
    |   X  |                                        X  |                  |
    |   I  A                                        I  A                  |
    |   S  X                                        S  X                  |
    |      I                                           I                  |
    |   D  S                                        D  S                  |
    |   E                                           E                     |
    |   S  S                                        S  S                  |
    |   C  C                                        C  C                  |
    |   R  A                                        R  A                  |
    |   I  L                                        I  L                  |
    |   P  E                                        P  E                  |
    |   T  S                                        T  S                  |
    |   I                                           I                     |
    |   O                                           O                     |
    |   N     X-AXIS SCALES                         N                     |
    |                                                                     |
    |       X-AXIS DESCRIPTION                                            |
    |                                                                     |
    |       "Plot Subsubtitle"                       "Legend Subinfo"     |
    |                                                                     |
    |                                                                     |
    |                                                                     |
    |_____________________________________________________________________|
 
 
    Now, as you might have gathered from the sketchs, the main problem
    we're addressing here is text --- gotta have text with the images
    we produce --- lots and lots of text.  Think of the situation this
    way, we make an image of something very, very interesting --- along
    with about three hundred other images which aren't so interesting
    because we're fiddling around with parameters.
 
    We dump everything to the printer and, three weeks later, we recognize
    that that one interesting plot contains some behavior we need to
    look at more closely.  Well, if we don't have the right stuff on the
    picture, we're screwed.  Maybe we can replicate parameters, maybe the
    number of possible combinations of parameters makes the problem far
    worse than producing all three hundred again and examining them one by
    one.
 
    Human labelling?  Hmmm.  Have humans do repetitive and tedious work
    instead of giving the task to the machine?  Yeah, right.
 
    Machine storage?  Well, these plots aren't likely to be so interesting
    that we're willing to store thousands of moderate-to-large postscript
    files indefinitely.  Print 'em and delete 'em --- as long as we can
    recreate the experiment from the picture (we'll probably only want
    a few dozen kept on-line at any given time).
 
    Nope, gotta have lots and lots of stuff right on the picture.  In
    fact, we *should* worry about what to put on the picture as we make
    the picture.  Each time we produce an image, we've already figured
    *how* to get that image --- there's no excuse for not linking an
    an approximation of *how* with the output --- even when the output
    goes offline.  If describing *how* is impossible, all the other
    questions are probably useless.
 
    It goes without saying that if the output isn't reasonably formatted,
    no one can or will look at it.
 
    Anyway, that's why we made those nasty movable position things and
    it's part of the reason why the set_plot_text routines give back
    a pointer to the image we created.  The user can add some initial
    text, get back some character pointers to save, print the plot,
    change the strings using the saved character pointers or EVEN
    add more text, print it again, etc.  We want to encourage the
    annotation of plots.  Repeat the set_plot_title routine and it'll
    append the new string under the old.  Heck, you can put paragraphs
    of stuff.
 
    What's the down side?  Well, if you put much on the page, it'll
    start to look funny and the image itself will shrink or invert (and
    it's possible that the routine will abort).  We counter this by
    allowing a separate page with just the image, blown up to show all
    the detail.  In the same sense, the user can shift the legend info
    to a separate page.
 
 
*/
 
class standard_plot : public plot_basics
{
protected:
   char * plot_title;
   boolean plot_title_set_by_user;

/* Some standard scales --- these include pointers to movable positions */

   scale_info scale_x;
   scale_info scale_y;
   scale_info scale_z;

/* Some position and image control booleans */

   boolean legend_required;
   boolean same_page_legend;
   boolean big_picture_too;

/* Some movable positions */

   movable_position visable_area_horizontal_center;
   movable_position plot_image_horizontal_center;
   movable_position plot_image_vertical_center;
   movable_position legend_image_horizontal_center;
   movable_position legend_image_vertical_center;

   movable_position visable_area_top_edge;
   movable_position combined_images_top_edge;

   movable_position plot_image_top_edge;
   movable_position plot_image_bottom_edge;
   movable_position plot_scale_bottom_edge;
   movable_position plot_label_bottom_edge;
   movable_position plot_desc_bottom_edge;
   movable_position plot_desc_text_boundary;
   movable_position plot_text_bottom_edge;

   movable_position legend_image_top_edge;
   movable_position legend_image_bottom_edge;
   movable_position legend_subinfo_position;

   movable_position combined_images_bottom_edge;
   movable_position combined_text_bottom_edge;
   movable_position visable_area_bottom_edge;

   movable_position visable_area_left_edge;
   movable_position plot_desc_left_edge;
   movable_position plot_label_left_edge;
   movable_position plot_scale_left_edge;
   movable_position plot_image_left_edge;
   movable_position plot_image_right_edge;
   movable_position plot_legend_separator;
   movable_position legend_desc_left_edge;
   movable_position legend_label_left_edge;
   movable_position legend_scale_left_edge;
   movable_position legend_image_left_edge;
   movable_position legend_image_right_edge;
   movable_position visable_area_right_edge;
 
   void set_position_links()
   {
      set_position_at_midpoint(visable_area_horizontal_center,
                               & visable_area_left_edge,
                               & visable_area_right_edge);
 
/* From top to bottom -- more or less */
      set_position_fixed(visable_area_top_edge,
                         PAGE_HEIGHT - TOP_MARGIN);
      set_position_link_to(combined_images_top_edge,
                           & visable_area_top_edge);
 
      set_position_link_to(plot_image_top_edge,
                           & combined_images_top_edge);
      set_position_at_midpoint(plot_image_vertical_center,
                               & plot_image_top_edge,
                               & plot_image_bottom_edge);
      set_position_link_to(plot_image_bottom_edge,
                           & plot_scale_bottom_edge);
      setup_box_location(scale_x, & plot_image_left_edge,
                                  & plot_image_right_edge,
                                  & plot_label_bottom_edge,
                                  & plot_scale_bottom_edge);
      set_positions_overlap_link(plot_label_bottom_edge,
                                 & plot_desc_bottom_edge,
                                 & plot_desc_text_boundary);
      set_positions_overlap_link(plot_desc_text_boundary,
                                & plot_text_bottom_edge,
                                & combined_images_bottom_edge);
 
      set_position_link_to(legend_image_top_edge,
                           & combined_images_top_edge);
      set_position_at_midpoint(legend_image_vertical_center,
                               & legend_image_top_edge,
                               & legend_image_bottom_edge);
      set_positions_overlap_link(legend_image_bottom_edge,
                                & legend_subinfo_position,
                                & combined_images_bottom_edge);
 
      set_positions_overlap_link(combined_images_bottom_edge,
                                & combined_text_bottom_edge,
                                & visable_area_bottom_edge);
      set_position_fixed(visable_area_bottom_edge, BOTTOM_MARGIN);
 
/* From left to right --- more or less */
      set_position_fixed(visable_area_left_edge,
                        LEFT_MARGIN);
      set_position_link_to(plot_desc_left_edge,
                           & visable_area_left_edge);
      set_position_link_to(plot_label_left_edge,
                           & plot_desc_left_edge);
      setup_box_location(scale_y, & plot_image_bottom_edge,
                                  & plot_image_top_edge,
                                  & plot_label_left_edge,
                                  & plot_scale_left_edge);
      set_position_link_to(plot_image_left_edge,
                           & plot_scale_left_edge);
      set_position_at_midpoint(plot_image_horizontal_center,
                               & plot_image_left_edge,
                               & plot_image_right_edge);
      set_position_relative_to(plot_image_right_edge,
                               & plot_legend_separator,
                               - 0.5 * INTERNAL_MARGIN);
      set_position_fixed(plot_legend_separator,
                         PAGE_WIDTH - RIGHT_MARGIN - LEGEND_SIZE
                         - 0.5 * INTERNAL_MARGIN);
      set_position_relative_to(legend_desc_left_edge,
                               & plot_legend_separator,
                               0.5 * INTERNAL_MARGIN);
      set_position_link_to(legend_label_left_edge,
                           & legend_desc_left_edge);
      setup_box_location(scale_z, & legend_image_bottom_edge,
                                  & legend_image_top_edge,
                                  & legend_label_left_edge,
                                  & legend_scale_left_edge);
      set_position_link_to(legend_image_left_edge,
                           & legend_scale_left_edge);
      set_position_at_midpoint(legend_image_horizontal_center,
                               & legend_image_left_edge,
                               & legend_image_right_edge);
      set_position_link_to(legend_image_left_edge,
                           & legend_scale_left_edge);
      set_position_link_to(legend_image_right_edge,
                           & visable_area_right_edge);
      set_position_fixed(visable_area_right_edge,
                         PAGE_WIDTH - RIGHT_MARGIN);
   };
 
/* Let the user play with margins and position controls */
public:
 
   void set_top_margin(double inches)
   {
      set_position_fixed(visable_area_top_edge,
                         PAGE_HEIGHT - PIXELS_PER_INCH * inches);
   };
 
   void set_bottom_margin(double inches)
   {
      set_position_fixed(visable_area_bottom_edge,
                         PIXELS_PER_INCH * inches);
   };
 
   void set_left_margin(double inches)
   {
      set_position_fixed(visable_area_left_edge,
                         PIXELS_PER_INCH * inches);
   };
 
   void set_right_margin(double inches)
   {
      set_position_fixed(visable_area_right_edge,
                         PAGE_WIDTH - PIXELS_PER_INCH * inches);
   };
 
   void reset_margin_defaults()
   {
      set_top_margin(TOP_MARGIN_INCHES);
      set_bottom_margin(LEFT_MARGIN_INCHES);
      set_left_margin(LEFT_MARGIN_INCHES);
      set_right_margin(RIGHT_MARGIN_INCHES);
   };
 
   void suppress_legend()
   {
      legend_required   = LOGICAL_FALSE;
      same_page_legend  = LOGICAL_FALSE;
      set_position_link_to(plot_image_right_edge,
                           & visable_area_right_edge);
   };
 
   void request_same_page_legend()
   {
      same_page_legend   = LOGICAL_TRUE;
      set_position_relative_to(plot_image_right_edge,
                               & plot_legend_separator,
                               - 0.5 * INTERNAL_MARGIN);
      set_position_relative_to(legend_desc_left_edge,
                               & plot_legend_separator,
                               0.5 * INTERNAL_MARGIN);
   };
 
   void request_separate_legend()
   {
      same_page_legend   = LOGICAL_FALSE;
      set_position_link_to(plot_image_right_edge,
                           & visable_area_right_edge);
      set_position_link_to(legend_desc_left_edge,
                           & visable_area_left_edge);
   };
 
   void request_big_picture_too()
   {
      big_picture_too = LOGICAL_TRUE;
   };
 
   void suppress_big_picture()
   {
      big_picture_too = LOGICAL_FALSE;
   };

protected:
 
/* Minor booleans to "watch" user access */
 
   boolean x_range_set;
   boolean y_range_set;
   boolean z_range_set;
 
/* allow end_user to set scales */
 
public:
   void set_x_range(UInt64 user_min, UInt64 user_max)
   {
      setup_range(scale_x, user_min, user_max);
      x_range_set = LOGICAL_TRUE;
   };
 
   void set_y_range(UInt64 user_min, UInt64 user_max)
   {
      setup_range(scale_y, user_min, user_max);
      y_range_set = LOGICAL_TRUE;
   };
 
/* If we have a z scale, we don't want the user to mess with it */
protected:
   void set_z_range(UInt64 user_max)
   {
      setup_range_force_storage_max(scale_z, 0, user_max);
      z_range_set = LOGICAL_TRUE;
   };
 
/* allow end_user to mess with scale options */
 
public:
   void suppress_x_axis_labels()
   {
      scale_x.labels_requested     = LOGICAL_FALSE;
      verify_scale_area(scale_x);
   };
   void enable_x_axis_labels()
   {
      scale_x.labels_requested     = LOGICAL_TRUE;
      verify_scale_area(scale_x);
   };
   void suppress_x_axis_tics()
   {
      scale_x.tics_requested       = LOGICAL_FALSE;
      verify_scale_area(scale_x);
   };
   void enable_x_axis_tics()
   {
      scale_x.tics_requested       = LOGICAL_TRUE;
      verify_scale_area(scale_x);
   };
   void suppress_x_axis_minor_tics()
   {
      scale_x.tics_requested       = LOGICAL_TRUE;
      scale_x.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_x);
   };
   void enable_x_axis_minor_tics()
   {
      scale_x.tics_requested       = LOGICAL_TRUE;
      scale_x.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_x);
   };
 
   void suppress_y_axis_labels()
   {
      scale_y.labels_requested     = LOGICAL_FALSE;
      verify_scale_area(scale_y);
   };
   void enable_y_axis_labels()
   {
      scale_y.labels_requested     = LOGICAL_TRUE;
      verify_scale_area(scale_y);
   };
   void suppress_y_axis_tics()
   {
      scale_y.tics_requested       = LOGICAL_FALSE;
      verify_scale_area(scale_y);
   };
   void enable_y_axis_tics()
   {
      scale_y.tics_requested       = LOGICAL_TRUE;
      verify_scale_area(scale_y);
   };
   void suppress_y_axis_minor_tics()
   {
      scale_y.tics_requested       = LOGICAL_TRUE;
      scale_y.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_y);
   };
   void enable_y_axis_minor_tics()
   {
      scale_y.tics_requested       = LOGICAL_TRUE;
      scale_y.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_y);
   };
 
   void suppress_z_axis_labels()
   {
      scale_z.labels_requested     = LOGICAL_FALSE;
      verify_scale_area(scale_z);
   };
   void enable_z_axis_labels()
   {
      scale_z.labels_requested     = LOGICAL_TRUE;
      verify_scale_area(scale_z);
   };
   void suppress_z_axis_tics()
   {
      scale_z.tics_requested       = LOGICAL_FALSE;
      verify_scale_area(scale_z);
   };
   void enable_z_axis_tics()
   {
      scale_z.tics_requested       = LOGICAL_TRUE;
      verify_scale_area(scale_z);
   };
   void suppress_z_axis_minor_tics()
   {
      scale_z.tics_requested       = LOGICAL_TRUE;
      scale_z.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_z);
   };
   void enable_z_axis_minor_tics()
   {
      scale_z.tics_requested       = LOGICAL_TRUE;
      scale_z.minor_tics_requested = LOGICAL_FALSE;
      verify_scale_area(scale_z);
   };
   void suppress_all_tics()
   {
      suppress_x_axis_tics();
      suppress_y_axis_tics();
      suppress_z_axis_tics();
   };
   void enable_all_tics()
   {
      enable_x_axis_tics();
      enable_y_axis_tics();
      enable_z_axis_tics();
   };
   void suppress_all_minor_tics()
   {
      suppress_x_axis_minor_tics();
      suppress_y_axis_minor_tics();
      suppress_z_axis_minor_tics();
   };
   void enable_all_minor_tics()
   {
      enable_x_axis_minor_tics();
      enable_y_axis_minor_tics();
      enable_z_axis_minor_tics();
   };
   void set_x_axis_label_font(const char * font_name,
                              double font_size)
   {
      free(scale_x.font_name);
      scale_x.font_name = frail_strdup(font_name);
      scale_x.font_size = font_size;
 
   };
   void set_x_axis_label_style(value_format style_request)
   {
      scale_x.style = style_request;
   };
   void set_y_axis_label_font(const char * font_name,
                              double font_size)
   {
      free(scale_y.font_name);
      scale_y.font_name = frail_strdup(font_name);
      scale_y.font_size = font_size;
 
   };
   void set_y_axis_label_style(value_format style_request)
   {
      scale_y.style = style_request;
   };
   void set_z_axis_label_font(const char * font_name,
                              double font_size)
   {
      free(scale_y.font_name);
      scale_y.font_name = frail_strdup(font_name);
      scale_y.font_size = font_size;
 
   };
   void set_z_axis_label_style(value_format style_request)
   {
      scale_z.style = style_request;
   };
 
/* Let the user play with text */
 
   char ** set_plot_title(const char * text_string,
                          const char * font_name,
                          double       font_size)
   {
      if (plot_title_set_by_user == LOGICAL_FALSE)
      {
         free(plot_title);
         plot_title = frail_strdup(text_string);
         plot_title_set_by_user = LOGICAL_TRUE;
      };
      coordinate_pair temp;
      temp.x = visable_area_horizontal_center;
      squeeze_before(temp.y, 
                     - FONT_FRACTION * font_size, 
                     - INTERLINE_FRACTION * font_size,
                     combined_images_top_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                ALL_IMAGES,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_footnote(const char * text_string,
                        const char * font_name,
                        double       font_size)
   {
      coordinate_pair temp;
      temp.x = visable_area_horizontal_center;
      squeeze_before_overlap(temp.y, 
                             - FONT_FRACTION * font_size,
                             - INTERLINE_FRACTION * font_size, 
                             combined_text_bottom_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                PLOT_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
 
   char ** set_plot_subtitle(const char * text_string,
                             const char * font_name,
                             double       font_size)
   {
      coordinate_pair temp;
      temp.x = plot_image_horizontal_center;
      squeeze_before(temp.y, 
                     - FONT_FRACTION * font_size, 
                     - INTERLINE_FRACTION * font_size,
                     plot_image_top_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                PLOT_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_plot_subsubtitle(const char * text_string,
                                const char * font_name,
                                double       font_size)
   {
      coordinate_pair temp;
      temp.x = plot_image_horizontal_center;
      squeeze_before_overlap(temp.y, 
                             - FONT_FRACTION * font_size,
                             - INTERLINE_FRACTION * font_size, 
                             plot_text_bottom_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                PLOT_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_legend_info(const char * text_string,
                           const char * font_name,
                           double       font_size)
   {
      coordinate_pair temp;
      temp.x = legend_image_horizontal_center;
      squeeze_before(temp.y, 
                     - FONT_FRACTION * font_size, 
                     - INTERLINE_FRACTION * font_size,
                     legend_image_top_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                LEGEND_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_legend_subinfo(const char * text_string,
                              const char * font_name,
                              double       font_size)
   {
      coordinate_pair temp;
      temp.x = legend_image_horizontal_center;
      squeeze_before_overlap(temp.y, 
                             - FONT_FRACTION * font_size,
                             - INTERLINE_FRACTION * font_size, 
                             legend_subinfo_position);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                LEGEND_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_x_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   {
      coordinate_pair temp;
      temp.x = plot_image_horizontal_center;
      squeeze_before_overlap(temp.y, 
                             - FONT_FRACTION * font_size,
                             - INTERLINE_FRACTION * font_size, 
                             plot_desc_bottom_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                DO_NOT_ROTATE,
                                                CENTER,
                                                PLOT_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_y_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   {
      coordinate_pair temp;
      temp.y = plot_image_vertical_center;
      squeeze_before(temp.x, 
                     FONT_FRACTION * font_size, 
                     2 * INTERLINE_FRACTION * font_size,
                     plot_desc_left_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                ROTATE_NINETY,
                                                CENTER,
                                                PLOT_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
   char ** set_z_axis_description(const char * text_string,
                                  const char * font_name,
                                  double       font_size)
   {
      coordinate_pair temp;
      temp.y = legend_image_vertical_center;
      squeeze_before(temp.x, 
                     FONT_FRACTION * font_size, 
                     2 * INTERLINE_FRACTION * font_size,
                     legend_desc_left_edge);
      plot_text * new_plot_text = set_plot_text(text_string,
                                                font_name,
                                                font_size,
                                                ROTATE_NINETY,
                                                CENTER,
                                                LEGEND_IMAGE,
                                                temp);
      return (& new_plot_text->text_string);
   };
 
/* Let the user access the above without thinking about fonts */
   char ** set_plot_title(const char * text_string)
   {
      return set_plot_title(text_string, "Times-Roman", 14);
   };
   char ** set_footnote(const char * text_string)
   {
      return set_footnote(text_string, "Times-Roman", 10);
   };
   char ** set_plot_subtitle(const char * text_string)
   {
      return set_plot_subtitle(text_string, "Times-Roman", 12);
   };
   char ** set_plot_subsubtitle(const char * text_string)
   {
      return set_plot_subsubtitle(text_string, "Times-Roman", 12);
   };
   char ** set_x_axis_description(const char * x_axis_description)
   {
      return set_x_axis_description(x_axis_description, "Courier", 10);
   };
   char ** set_y_axis_description(const char * y_axis_description)
   {
      return set_y_axis_description(y_axis_description, "Courier", 10);
   };
   char ** set_z_axis_description(const char * z_axis_description)
   {
      return set_z_axis_description(z_axis_description, "Courier", 10);
   };
   char ** set_legend_info(const char * legend_info)
   {
      return set_legend_info(legend_info, "Times-Roman", 12);
   };
   char ** set_legend_subinfo(const char * legend_subinfo)
   {
      return set_legend_subinfo(legend_subinfo, "Times-Roman", 12);
   };
 
private:
   void byte_print(FILE * output_stream, UChar byte_data)
   {
      static const char hex_digit[]   = "0123456789abcdef";
#ifdef INVERT_IMAGE
      byte_data = 255 - byte_data;
#endif
      fprintf(output_stream, "%c%c ", hex_digit[byte_data >> 4],
                                      hex_digit[byte_data & 15]);
   };
   void emit_defs(FILE * output_stream)
   {
      fprintf(output_stream, "/cshow {\n");
      fprintf(output_stream, "    dup stringwidth pop 2 div neg 0 rmoveto show\n");
      fprintf(output_stream, "} bind def\n\n");
      fprintf(output_stream, "/rshow {\n");
      fprintf(output_stream, "    dup stringwidth pop neg 0 rmoveto show\n");
      fprintf(output_stream, "} bind def\n\n");
      emit_linetone_def(output_stream);
   };
 
 
   void emit_linetone_def(FILE * output_stream)
   {
      fprintf(output_stream, "/linedata %d string def\n", scale_x.storage_max);
      fprintf(output_stream, "gsave\n" );
      fprintf(output_stream, "/linetone <<\n" );
      fprintf(output_stream, "   /HalftoneType 3\n" );
      fprintf(output_stream, "   /Width 9\n" );
      fprintf(output_stream, "   /Height 9\n" );
      fprintf(output_stream, "   /Thresholds <00 20 40 60 80 9f bf df ff\n" );
      fprintf(output_stream, "                ff 00 20 40 60 80 9f bf df\n" );
      fprintf(output_stream, "                df ff 00 20 40 60 80 9f bf\n" );
      fprintf(output_stream, "                bf df ff 00 20 40 60 80 9f\n" );
      fprintf(output_stream, "                9f bf df ff 00 20 40 60 80\n" );
      fprintf(output_stream, "                80 9f bf df ff 00 20 40 60\n" );
      fprintf(output_stream, "                60 80 9f bf df ff 00 20 40\n" );
      fprintf(output_stream, "                40 60 80 9f bf df ff 00 20\n" );
      fprintf(output_stream, "                20 40 60 80 9f bf df ff 00>\n" );
      fprintf(output_stream, "    /TransferFunction { 1 exch sub }\n" );
      fprintf(output_stream, ">> def\n" );
   };
 
protected:
   virtual void emit_image(FILE * output_stream,
                           double left_x_position,
                           double right_x_position,
                           double upper_y_position,
                           double lower_y_position) = 0;
 
   virtual void emit_legend_image(FILE * output_stream,
                                  double left_x_position,
                                  double right_x_position,
                                  double upper_y_position,
                                  double lower_y_position)
   {
      double size_x = right_x_position - left_x_position;
      double size_y = upper_y_position - lower_y_position;
      fprintf(output_stream, "%% Legend image\n");
      fprintf(output_stream, "gsave %g %g translate\n",
                             left_x_position,
                             lower_y_position);
      fprintf(output_stream, "      1 1 scale\n");
      if (scale_z.user_max < 255)
      {
         fprintf(output_stream, "      1 %u 8\n", scale_z.user_max + 1);
         fprintf(output_stream, "      [ %g 0 0 %g 0 0 ]\n",
                                 1.0 / size_x,
                                 (scale_z.user_max + 1.0) / size_y);
         fprintf(output_stream, "      {<\n      ");
 
         UInt32 line_data_count = 0;
         for (UChar loop = 0; loop <= scale_z.user_max; loop++)
         {
            if (line_data_count == 20)
            {
               fprintf(output_stream, "\n      ");
               line_data_count = 0;
            };
            byte_print(output_stream, (UChar)storage_location(scale_z, loop));
            line_data_count++;
         };
         fprintf(output_stream, "      >}\n");
      }
      else
      {
         fprintf(output_stream, "      1 256 8\n");
         fprintf(output_stream, "      [ %g 0 0 %g 0 0 ]\n",
                                 1.0 / size_x, 256.0 / size_y);
         fprintf(output_stream, "      {<\n      ");
 
         UInt32 line_data_count = 0;
         for (UInt32 loop = 0; loop < 256; loop++)
         {
            if (line_data_count == 16)
            {
               fprintf(output_stream, "\n      ");
               line_data_count = 0;
            };
            byte_print(output_stream, (UChar)(loop));
         };
         fprintf(output_stream, "      >}\n");
      };
      fprintf(output_stream, "      image\n");
      fprintf(output_stream, "grestore\n");
   };
 
   void emit_plot(FILE * output_stream)
   {
      time_t now;
      time(&now);
      UInt32 page_count = 1;
 
      fprintf(output_stream, "%%!PS-Adobe-3.0" );
      fprintf(output_stream, "\n%%%%BoundingBox: %g %g %g %g\n",
                             0.0, 0.0,
                             (double)PAGE_WIDTH, (double)PAGE_HEIGHT);
      fprintf(output_stream, "%%%%Creator:  %s\n", VERSION_NAME);
      if (getenv("USER"))
         fprintf(output_stream, "%%%%For: %s\n", getenv("USER") );
      fprintf(output_stream, "%%%%Title: %s\n", plot_title);
      fprintf(output_stream, "%%%%CreationDate: %s", ctime(&now));
      emit_distinct_fonts(output_stream);
      fprintf(output_stream, "%%%%Orientation: Landscape\n");
      fprintf(output_stream, "%%%%EndComments\n" );
      fprintf(output_stream, "30 dict begin /vmstate save def\n" );
      emit_defs(output_stream);
 
      fprintf(output_stream, "%%%%Page: 1 1\n");
      fprintf(output_stream, "90 rotate 0 %g translate %% landscape\n",
                             -PAGE_HEIGHT);
      if ((legend_required == LOGICAL_TRUE) && (same_page_legend == LOGICAL_TRUE))
      {
         emit_plot_text(output_stream, ALL_IMAGES);
      }
      else
      {
         emit_plot_text(output_stream, PLOT_IMAGE);
      };
      emit_horizontal_scale(output_stream, scale_x);
      emit_vertical_scale(output_stream, scale_y);
      emit_outline_box(output_stream,
                       position(plot_image_left_edge),
                       position(plot_image_right_edge),
                       position(plot_image_top_edge),
                       position(plot_image_bottom_edge));
      emit_image(output_stream,
                       position(plot_image_left_edge),
                       position(plot_image_right_edge),
                       position(plot_image_top_edge),
                       position(plot_image_bottom_edge));
      if (legend_required == LOGICAL_TRUE)
      {
         if (same_page_legend == LOGICAL_FALSE)
         {
            fprintf(output_stream, "showpage\n");
            page_count++;
            fprintf(output_stream, "%%%%Page: %d %d\n",
                                   page_count, page_count);
            emit_plot_text(output_stream, LEGEND_IMAGE);
         };
         emit_vertical_scale(output_stream, scale_z);
         emit_outline_box(output_stream,
                          position(legend_image_left_edge),
                          position(legend_image_right_edge),
                          position(legend_image_top_edge),
                          position(legend_image_bottom_edge));
         emit_legend_image(output_stream,
                          position(legend_image_left_edge),
                          position(legend_image_right_edge),
                          position(legend_image_top_edge),
                          position(legend_image_bottom_edge));
      };
 
      if (big_picture_too)
      {
         fprintf(output_stream, "showpage\n");
         page_count++;
         fprintf(output_stream, "%%%%Page: %d %d\n",
                                page_count, page_count);
         fprintf(output_stream, "90 rotate 0 %g translate %% landscape\n",
                                -PAGE_HEIGHT);
         movable_position save_plot_image_left_edge   = plot_image_left_edge;
         movable_position save_plot_image_right_edge  = plot_image_right_edge;
         movable_position save_plot_image_top_edge    = plot_image_top_edge;
         movable_position save_plot_image_bottom_edge = plot_image_bottom_edge;
         plot_image_left_edge   = visable_area_left_edge;
         plot_image_right_edge  = visable_area_right_edge;
         plot_image_top_edge    = visable_area_top_edge;
         plot_image_bottom_edge = visable_area_bottom_edge;
         emit_outline_box(output_stream,
                          position(plot_image_left_edge),
                          position(plot_image_right_edge),
                          position(plot_image_top_edge),
                          position(plot_image_bottom_edge));
         emit_image(output_stream,
                       position(visable_area_left_edge),
                       position(visable_area_right_edge),
                       position(visable_area_top_edge),
                       position(visable_area_bottom_edge));
         plot_image_left_edge   = save_plot_image_left_edge;
         plot_image_right_edge  = save_plot_image_right_edge;
         plot_image_top_edge    = save_plot_image_top_edge;
         plot_image_bottom_edge = save_plot_image_bottom_edge;
      };
      fprintf(output_stream, "showpage\n");
      fprintf(output_stream, "vmstate restore end\n" );
      fprintf(output_stream, "%%%%Trailer\n" );
   };
 
/*
   For some types of plots, we can adjust paramters after printing and
   print again, any class using this as a base class would need access
   to adjust internals just after the options have been set anf just
   before printing begins.  By constrast, some larger classes will
   need access after the options has been set, but before allowing
   the user access to other public manipulations of the image ---
   the options_complete routine is intended for that purpose */
 
   virtual void prepare_print() = 0;
 
public:
   virtual void options_complete() = 0;
 
/* normally, the following would be public, but we don't want the end
   user getting it accidently */
protected:
   standard_plot()
   {
      plot_title = frail_strdup("Untitled");
      plot_title_set_by_user = LOGICAL_FALSE;
      setup_scale_parameters(scale_x, (UInt64)PAGE_WIDTH,  PLOT_IMAGE);
      setup_scale_parameters(scale_y, (UInt64)PAGE_HEIGHT, PLOT_IMAGE);
      setup_scale_parameters(scale_z, 255,                 LEGEND_IMAGE);
      set_position_links();
      request_same_page_legend();
      big_picture_too = LOGICAL_FALSE;
   };
 
public:
   void print_plot(FILE * output_stream)
   {
      prepare_print();
      verify_scale_area(scale_x);
      verify_scale_area(scale_y);
      verify_scale_area(scale_z);
      emit_plot(output_stream);
   };
 
protected:
   void kill_standard_plot()
   {
      kill_plot_basics();
      free(scale_x.font_name);
      free(scale_y.font_name);
      free(scale_z.font_name);
   };
   friend density_plot;
   friend histogram_plot;
};
 
#endif

